from enum import IntFlag

from wacky.basic_types.integers import u32
from wacky.enums.adapter import enum_adapter


class PackageFlag(IntFlag):
    NewlyCreated = 0x00000001
    ClientOptional = 0x00000002
    ServerSideOnly = 0x00000004
    # Unused = 0x00000008
    CompiledIn = 0x00000010
    ForDiffing = 0x00000020
    EditorOnly = 0x00000040
    Developer = 0x00000080
    # Loaded only in uncooked builds (i.e. runtime in editor)
    UncookedOnly = 0x00000100
    Cooked = 0x00000200
    # Package doesn't contain any asset object (although asset tags can be
    # present)
    ContainsNoAsset = 0x00000400
    # Unused = 0x00000800
    # Unused = 0x00001000
    # Uses unversioned property serialization instead of versioned tagged
    # property serialization
    UnversionedProperties = 0x00002000
    # Contains map data (UObjects only referenced by a single ULevel) but is
    # stored in a different package
    ContainsMapData = 0x00004000
    # Unused = 0x00008000
    Compiling = 0x00010000
    # Package contains a ULevel/ UWorld object
    ContainsMap = 0x00020000
    # Package contains any data to be gathered by localization
    RequiresLocalizationGather = 0x00040000
    # Unused = 0x00080000
    # Package was created for the purpose of PIE
    PlayInEditor = 0x00100000
    # Package is allowed to contain UClass objects
    ContainsScript = 0x00200000
    # Editor should not export asset in this package
    DisallowExport = 0x00400000
    # Unused = 0x00800000
    # Unused = 0x01000000
    # Unused = 0x02000000
    # Unused = 0x04000000
    # Unused = 0x08000000
    # This package should resolve dynamic imports from its export at runtime.
    DynamicImports = 0x10000000
    # This package contains elements that are runtime generated, and may not
    # follow standard loading order rules
    RuntimeGenerated = 0x20000000
    # This package is reloading in the cooker, try to avoid getting data we
    # will never need. We won't save this package.
    ReloadingForCooker = 0x40000000
    # Package has editor-only data filtered out
    FilterEditorOnly = 0x80000000


PackageFlagStruct = enum_adapter(u32, PackageFlag)



from dataclasses import dataclass
from typing import Optional

from construct import Adapter, CString, If, IfThenElse
from construct_typed import DataclassMixin, DataclassStruct, csfield

from wacky.basic_types.integers import i32


@dataclass
class RawString(DataclassMixin):
    signed_length: int = csfield(i32)
    value: Optional[str] = csfield(
        If(
            lambda this: this.signed_length != 0,
            IfThenElse(
                lambda this: this.signed_length < 0,
                CString("utf-16-le"),
                CString("ascii"),
            ),
        )
    )


class StringAdapter(Adapter):
    def _decode(self, obj: RawString, context, path: str) -> str:
        return obj.value or ""

    def _encode(self, obj: str, context, path) -> RawString:
        if not obj:
            return RawString(signed_length=0, value=obj)
        elif obj.isascii():
            return RawString(signed_length=len(obj) + 1, value=obj)
        else:
            return RawString(signed_length=-(len(obj) + 1), value=obj)


String = StringAdapter(DataclassStruct(RawString))



from construct import Adapter

from wacky.basic_types.integers import i32


class LargeBoolAdapter(Adapter):
    def _decode(self, obj: int, context, path: str) -> bool:
        return obj != 0

    def _encode(self, obj: bool, context, path) -> int:
        return int(obj)


LargeBool = LargeBoolAdapter(i32)



from dataclasses import dataclass
from enum import Enum, auto
from typing import Optional

from construct_typed import DataclassMixin, DataclassStruct, csfield

from wacky.basic_types.integers import i32


class IndexKind(Enum):
    NULL = auto()
    IMPORT = auto()
    EXPORT = auto()


@dataclass
class PackageIndex(DataclassMixin):
    signed_value: int = csfield(i32)

    @property
    def kind(self) -> IndexKind:
        if self.signed_value == 0:
            return IndexKind.NULL
        elif self.signed_value < 0:
            return IndexKind.IMPORT
        else:
            return IndexKind.EXPORT

    @property
    def value(self) -> Optional[int]:
        kind = self.kind
        if kind == IndexKind.NULL:
            return None
        elif kind == IndexKind.IMPORT:
            return -self.signed_value - 1
        else:
            return self.signed_value - 1

    def jsonify(self) -> int:
        return self.signed_value

    def update(self, new: int):
        self.signed_value = new


PackageIndexStruct = DataclassStruct(PackageIndex)



from typing import BinaryIO

from wacky.package import Package, PackageStruct


def load(uasset: BinaryIO, uexp: BinaryIO) -> Package:
    bytes_ = uasset.read() + uexp.read()
    return PackageStruct.parse(bytes_)


def dump(package: Package, uasset: BinaryIO, uexp: BinaryIO) -> None:
    """Dump package to uasset and uexp"""
    bytes_with_wrong_offsets = PackageStruct.build(package)
    reparsed = PackageStruct.parse(bytes_with_wrong_offsets)
    reparsed.fix_offsets_and_sizes()
    bytes_ = PackageStruct.build(reparsed)

    uasset_bytes = bytes_[: package.total_header_size]
    uexp_bytes = bytes_[package.total_header_size :]
    uasset.write(uasset_bytes)
    uexp.write(uexp_bytes)


def jsonify(uasset: BinaryIO, uexp: BinaryIO) -> list:
    package = load(uasset=uasset, uexp=uexp)
    return package.jsonify()


def update(uasset: BinaryIO, uexp: BinaryIO, changes: list) -> Package:
    package = load(uasset=uasset, uexp=uexp)
    package.update(changes)
    return package



from functools import reduce
from typing import List
from uuid import UUID

from construct import Adapter, Array

from wacky.basic_types.integers import u32

RawGUID = Array(4, u32)


class GUIDAdapter(Adapter):
    def _decode(self, obj: List[int], context, path: str) -> UUID:
        value = reduce(lambda x, y: (x << 32) + y, obj)
        return UUID(int=value)

    def _encode(self, obj: UUID, context, path) -> List[int]:
        return [
            int(obj.hex[:8], 16),
            int(obj.hex[8:16], 16),
            int(obj.hex[16:24], 16),
            int(obj.hex[24:], 16),
        ]


GUIDStruct = GUIDAdapter(RawGUID)



from dataclasses import dataclass
from enum import IntEnum
from typing import List, Type, Union

from construct import LazyBound, PrefixedArray, Switch
from construct_typed import DataclassMixin, DataclassStruct, csfield

from wacky.basic_types.integers import i32, u32
from wacky.basic_types.large_bool import LargeBool
from wacky.basic_types.string import String
from wacky.enums.adapter import enum_adapter


class ValueType(IntEnum):
    NONE = 0
    BOOLEAN = 1
    STRING = 2
    ARRAY = 3
    OBJECT = 4


ValueTypeStruct = enum_adapter(i32, ValueType)


@dataclass
class LocMetadataValue(DataclassMixin):
    value_type: ValueType = csfield(ValueTypeStruct)
    value: Union[bool, str, List["LocMetadataValue"]] = csfield(
        Switch(
            lambda this: this.value_type,
            {
                ValueType.BOOLEAN: LargeBool,
                ValueType.STRING: String,
                ValueType.ARRAY: PrefixedArray(u32, LazyBound(get_loc_metadata_value)),
                # ValueType.OBJECT: Error  # LazyBound(lambda: LocMetadataObject),
            },
            # default=Error,
        )
    )


def get_loc_metadata_value() -> Type[DataclassMixin]:
    return LocMetadataValue


@dataclass
class LocMetadataEntry(DataclassMixin):
    key: str = csfield(String)
    value: LocMetadataValue = csfield(DataclassStruct(LocMetadataValue))


LocMetadataObject = PrefixedArray(u32, DataclassStruct(LocMetadataEntry))


@dataclass
class TextSourceSiteContext(DataclassMixin):
    key_name: str = csfield(String)
    site_description: str = csfield(String)
    is_editor_only: bool = csfield(LargeBool)
    is_optional: bool = csfield(LargeBool)
    info_metadata: List[LocMetadataEntry] = csfield(LocMetadataObject)
    key_metadata: List[LocMetadataEntry] = csfield(LocMetadataObject)


@dataclass
class GatherableText(DataclassMixin):
    namespace_name: str = csfield(String)
    source_string: str = csfield(String)
    source_string_metadata: List[LocMetadataEntry] = csfield(LocMetadataObject)
    source_site_contexts: List[TextSourceSiteContext] = csfield(
        PrefixedArray(u32, DataclassStruct(TextSourceSiteContext))
    )



from enum import IntEnum
from functools import partial
from typing import Type, TypeVar

from construct import Container, ExprAdapter, FormatField

T = TypeVar("T", bound=IntEnum)


def decode_int(enum_type: Type[T], obj: int, ctx: Container) -> T:
    return enum_type(obj)


def encode_enum(obj: T, ctx: Container) -> int:
    return obj.value


def enum_adapter(underlying_type: FormatField, enum_type: Type[IntEnum]) -> ExprAdapter:
    return ExprAdapter(
        subcon=underlying_type,
        decoder=partial(decode_int, enum_type),
        encoder=encode_enum,
    )



from enum import IntEnum, auto


class UE4Version(IntEnum):
    Cooked = 0  # not a real ObjectVersion
    OLDEST_LOADABLE_PACKAGE = 214

    BLUEPRINT_VARS_NOT_READ_ONLY = auto()
    STATIC_MESH_STORE_NAV_COLLISION = auto()
    ATMOSPHERIC_FOG_DECAY_NAME_CHANGE = auto()
    SCENECOMP_TRANSLATION_TO_LOCATION = auto()
    MATERIAL_ATTRIBUTES_REORDERING = auto()
    COLLISION_PROFILE_SETTING = auto()
    BLUEPRINT_SKEL_TEMPORARY_TRANSIENT = auto()
    BLUEPRINT_SKEL_SERIALIZED_AGAIN = auto()
    BLUEPRINT_SETS_REPLICATION = auto()
    WORLD_LEVEL_INFO = auto()
    AFTER_CAPSULE_HALF_HEIGHT_CHANGE = auto()
    ADDED_NAMESPACE_AND_KEY_DATA_TO_FTEXT = auto()
    ATTENUATION_SHAPES = auto()
    LIGHTCOMPONENT_USE_IES_TEXTURE_MULTIPLIER_ON_NON_IES_BRIGHTNESS = auto()
    REMOVE_INPUT_COMPONENTS_FROM_BLUEPRINTS = auto()
    VARK2NODE_USE_MEMBERREFSTRUCT = auto()
    REFACTOR_MATERIAL_EXPRESSION_SCENECOLOR_AND_SCENEDEPTH_INPUTS = auto()
    SPLINE_MESH_ORIENTATION = auto()
    REVERB_EFFECT_ASSET_TYPE = auto()
    MAX_TEXCOORD_INCREASED = auto()
    SPEEDTREE_STATICMESH = auto()
    LANDSCAPE_COMPONENT_LAZY_REFERENCES = auto()
    SWITCH_CALL_NODE_TO_USE_MEMBER_REFERENCE = auto()
    ADDED_SKELETON_ARCHIVER_REMOVAL = auto()
    ADDED_SKELETON_ARCHIVER_REMOVAL_SECOND_TIME = auto()
    BLUEPRINT_SKEL_CLASS_TRANSIENT_AGAIN = auto()
    ADD_COOKED_TO_UCLASS = auto()
    DEPRECATED_STATIC_MESH_THUMBNAIL_PROPERTIES_REMOVED = auto()
    COLLECTIONS_IN_SHADERMAPID = auto()
    REFACTOR_MOVEMENT_COMPONENT_HIERARCHY = auto()
    FIX_TERRAIN_LAYER_SWITCH_ORDER = auto()
    ALL_PROPS_TO_CONSTRAINTINSTANCE = auto()
    LOW_QUALITY_DIRECTIONAL_LIGHTMAPS = auto()
    ADDED_NOISE_EMITTER_COMPONENT = auto()
    ADD_TEXT_COMPONENT_VERTICAL_ALIGNMENT = auto()
    ADDED_FBX_ASSET_IMPORT_DATA = auto()
    REMOVE_LEVELBODYSETUP = auto()
    REFACTOR_CHARACTER_CROUCH = auto()
    SMALLER_DEBUG_MATERIALSHADER_UNIFORM_EXPRESSIONS = auto()
    APEX_CLOTH = auto()
    SAVE_COLLISIONRESPONSE_PER_CHANNEL = auto()
    ADDED_LANDSCAPE_SPLINE_EDITOR_MESH = auto()
    CHANGED_MATERIAL_REFACTION_TYPE = auto()
    REFACTOR_PROJECTILE_MOVEMENT = auto()
    REMOVE_PHYSICALMATERIALPROPERTY = auto()
    PURGED_FMATERIAL_COMPILE_OUTPUTS = auto()
    ADD_COOKED_TO_LANDSCAPE = auto()
    CONSUME_INPUT_PER_BIND = auto()
    SOUND_CLASS_GRAPH_EDITOR = auto()
    FIXUP_TERRAIN_LAYER_NODES = auto()
    RETROFIT_CLAMP_EXPRESSIONS_SWAP = auto()
    REMOVE_LIGHT_MOBILITY_CLASSES = auto()
    REFACTOR_PHYSICS_BLENDING = auto()
    WORLD_LEVEL_INFO_UPDATED = auto()
    STATIC_SKELETAL_MESH_SERIALIZATION_FIX = auto()
    REMOVE_STATICMESH_MOBILITY_CLASSES = auto()
    REFACTOR_PHYSICS_TRANSFORMS = auto()
    REMOVE_ZERO_TRIANGLE_SECTIONS = auto()
    CHARACTER_MOVEMENT_DECELERATION = auto()
    CAMERA_ACTOR_USING_CAMERA_COMPONENT = auto()
    CHARACTER_MOVEMENT_DEPRECATE_PITCH_ROLL = auto()
    REBUILD_TEXTURE_STREAMING_DATA_ON_LOAD = auto()
    SUPPORT_32BIT_STATIC_MESH_INDICES = auto()
    ADDED_CHUNKID_TO_ASSETDATA_AND_UPACKAGE = auto()
    CHARACTER_DEFAULT_MOVEMENT_BINDINGS = auto()
    APEX_CLOTH_LOD = auto()
    ATMOSPHERIC_FOG_CACHE_DATA = auto()
    VAR_UE4_ARRAY_PROPERTY_INNER_TAGS = auto()
    KEEP_SKEL_MESH_INDEX_DATA = auto()
    BODYSETUP_COLLISION_CONVERSION = auto()
    REFLECTION_CAPTURE_COOKING = auto()
    REMOVE_DYNAMIC_VOLUME_CLASSES = auto()
    STORE_HASCOOKEDDATA_FOR_BODYSETUP = auto()
    REFRACTION_BIAS_TO_REFRACTION_DEPTH_BIAS = auto()
    REMOVE_SKELETALPHYSICSACTOR = auto()
    PC_ROTATION_INPUT_REFACTOR = auto()
    LANDSCAPE_PLATFORMDATA_COOKING = auto()
    CREATEEXPORTS_CLASS_LINKING_FOR_BLUEPRINTS = auto()
    REMOVE_NATIVE_COMPONENTS_FROM_BLUEPRINT_SCS = auto()
    REMOVE_SINGLENODEINSTANCE = auto()
    CHARACTER_BRAKING_REFACTOR = auto()
    VOLUME_SAMPLE_LOW_QUALITY_SUPPORT = auto()
    SPLIT_TOUCH_AND_CLICK_ENABLES = auto()
    HEALTH_DEATH_REFACTOR = auto()
    SOUND_NODE_ENVELOPER_CURVE_CHANGE = auto()
    POINT_LIGHT_SOURCE_RADIUS = auto()
    SCENE_CAPTURE_CAMERA_CHANGE = auto()
    MOVE_SKELETALMESH_SHADOWCASTING = auto()
    CHANGE_SETARRAY_BYTECODE = auto()
    MATERIAL_INSTANCE_BASE_PROPERTY_OVERRIDES = auto()
    COMBINED_LIGHTMAP_TEXTURES = auto()
    BUMPED_MATERIAL_EXPORT_GUIDS = auto()
    BLUEPRINT_INPUT_BINDING_OVERRIDES = auto()
    FIXUP_BODYSETUP_INVALID_CONVEX_TRANSFORM = auto()
    FIXUP_STIFFNESS_AND_DAMPING_SCALE = auto()
    REFERENCE_SKELETON_REFACTOR = auto()
    K2NODE_REFERENCEGUIDS = auto()
    FIXUP_ROOTBONE_PARENT = auto()
    TEXT_RENDER_COMPONENTS_WORLD_SPACE_SIZING = auto()
    MATERIAL_INSTANCE_BASE_PROPERTY_OVERRIDES_PHASE_2 = auto()
    CLASS_NOTPLACEABLE_ADDED = auto()
    WORLD_LEVEL_INFO_LOD_LIST = auto()
    CHARACTER_MOVEMENT_VARIABLE_RENAMING_1 = auto()
    FSLATESOUND_CONVERSION = auto()
    WORLD_LEVEL_INFO_ZORDER = auto()
    PACKAGE_REQUIRES_LOCALIZATION_GATHER_FLAGGING = auto()
    BP_ACTOR_VARIABLE_DEFAULT_PREVENTING = auto()
    TEST_ANIMCOMP_CHANGE = auto()
    EDITORONLY_BLUEPRINTS = auto()
    EDGRAPHPINTYPE_SERIALIZATION = auto()
    NO_MIRROR_BRUSH_MODEL_COLLISION = auto()
    CHANGED_CHUNKID_TO_BE_AN_ARRAY_OF_CHUNKIDS = auto()
    WORLD_NAMED_AFTER_PACKAGE = auto()
    SKY_LIGHT_COMPONENT = auto()
    WORLD_LAYER_ENABLE_DISTANCE_STREAMING = auto()
    REMOVE_ZONES_FROM_MODEL = auto()
    FIX_ANIMATIONBASEPOSE_SERIALIZATION = auto()
    SUPPORT_8_BONE_INFLUENCES_SKELETAL_MESHES = auto()
    ADD_OVERRIDE_GRAVITY_FLAG = auto()
    SUPPORT_GPUSKINNING_8_BONE_INFLUENCES = auto()
    ANIM_SUPPORT_NONUNIFORM_SCALE_ANIMATION = auto()
    ENGINE_VERSION_OBJECT = auto()
    PUBLIC_WORLDS = auto()
    SKELETON_GUID_SERIALIZATION = auto()
    CHARACTER_MOVEMENT_WALKABLE_FLOOR_REFACTOR = auto()
    INVERSE_SQUARED_LIGHTS_DEFAULT = auto()
    DISABLED_SCRIPT_LIMIT_BYTECODE = auto()
    PRIVATE_REMOTE_ROLE = auto()
    FOLIAGE_STATIC_MOBILITY = auto()
    BUILD_SCALE_VECTOR = auto()
    FOLIAGE_COLLISION = auto()
    SKY_BENT_NORMAL = auto()
    LANDSCAPE_COLLISION_DATA_COOKING = auto()
    MORPHTARGET_CPU_TANGENTZDELTA_FORMATCHANGE = auto()
    SOFT_CONSTRAINTS_USE_MASS = auto()
    REFLECTION_DATA_IN_PACKAGES = auto()
    FOLIAGE_MOVABLE_MOBILITY = auto()
    UNDO_BREAK_MATERIALATTRIBUTES_CHANGE = auto()
    ADD_CUSTOMPROFILENAME_CHANGE = auto()
    FLIP_MATERIAL_COORDS = auto()
    MEMBERREFERENCE_IN_PINTYPE = auto()
    VEHICLES_UNIT_CHANGE = auto()
    ANIMATION_REMOVE_NANS = auto()
    SKELETON_ASSET_PROPERTY_TYPE_CHANGE = auto()
    FIX_BLUEPRINT_VARIABLE_FLAGS = auto()
    VEHICLES_UNIT_CHANGE2 = auto()
    UCLASS_SERIALIZE_INTERFACES_AFTER_LINKING = auto()
    STATIC_MESH_SCREEN_SIZE_LODS = auto()
    FIX_MATERIAL_COORDS = auto()
    SPEEDTREE_WIND_V7 = auto()
    LOAD_FOR_EDITOR_GAME = auto()
    SERIALIZE_RICH_CURVE_KEY = auto()
    MOVE_LANDSCAPE_MICS_AND_TEXTURES_WITHIN_LEVEL = auto()
    FTEXT_HISTORY = auto()
    FIX_MATERIAL_COMMENTS = auto()
    STORE_BONE_EXPORT_NAMES = auto()
    MESH_EMITTER_INITIAL_ORIENTATION_DISTRIBUTION = auto()
    DISALLOW_FOLIAGE_ON_BLUEPRINTS = auto()
    FIXUP_MOTOR_UNITS = auto()
    DEPRECATED_MOVEMENTCOMPONENT_MODIFIED_SPEEDS = auto()
    RENAME_CANBECHARACTERBASE = auto()
    GAMEPLAY_TAG_CONTAINER_TAG_TYPE_CHANGE = auto()
    FOLIAGE_SETTINGS_TYPE = auto()
    STATIC_SHADOW_DEPTH_MAPS = auto()
    ADD_TRANSACTIONAL_TO_DATA_ASSETS = auto()
    ADD_LB_WEIGHTBLEND = auto()
    ADD_ROOTCOMPONENT_TO_FOLIAGEACTOR = auto()
    FIX_MATERIAL_PROPERTY_OVERRIDE_SERIALIZE = auto()
    ADD_LINEAR_COLOR_SAMPLER = auto()
    ADD_STRING_ASSET_REFERENCES_MAP = auto()
    BLUEPRINT_USE_SCS_ROOTCOMPONENT_SCALE = auto()
    LEVEL_STREAMING_DRAW_COLOR_TYPE_CHANGE = auto()
    CLEAR_NOTIFY_TRIGGERS = auto()
    SKELETON_ADD_SMARTNAMES = auto()
    ADDED_CURRENCY_CODE_TO_FTEXT = auto()
    ENUM_CLASS_SUPPORT = auto()
    FIXUP_WIDGET_ANIMATION_CLASS = auto()
    SOUND_COMPRESSION_TYPE_ADDED = auto()
    AUTO_WELDING = auto()
    RENAME_CROUCHMOVESCHARACTERDOWN = auto()
    LIGHTMAP_MESH_BUILD_SETTINGS = auto()
    RENAME_SM3_TO_ES3_1 = auto()
    DEPRECATE_UMG_STYLE_ASSETS = auto()
    POST_DUPLICATE_NODE_GUID = auto()
    RENAME_CAMERA_COMPONENT_VIEW_ROTATION = auto()
    CASE_PRESERVING_FNAME = auto()
    RENAME_CAMERA_COMPONENT_CONTROL_ROTATION = auto()
    FIX_REFRACTION_INPUT_MASKING = auto()
    GLOBAL_EMITTER_SPAWN_RATE_SCALE = auto()
    CLEAN_DESTRUCTIBLE_SETTINGS = auto()
    CHARACTER_MOVEMENT_UPPER_IMPACT_BEHAVIOR = auto()
    BP_MATH_VECTOR_EQUALITY_USES_EPSILON = auto()
    FOLIAGE_STATIC_LIGHTING_SUPPORT = auto()
    SLATE_COMPOSITE_FONTS = auto()
    REMOVE_SAVEGAMESUMMARY = auto()

    REMOVE_SKELETALMESH_COMPONENT_BODYSETUP_SERIALIZATION = auto()
    SLATE_BULK_FONT_DATA = auto()
    ADD_PROJECTILE_FRICTION_BEHAVIOR = auto()
    MOVEMENTCOMPONENT_AXIS_SETTINGS = auto()
    GRAPH_INTERACTIVE_COMMENTBUBBLES = auto()
    LANDSCAPE_SERIALIZE_PHYSICS_MATERIALS = auto()
    RENAME_WIDGET_VISIBILITY = auto()
    ANIMATION_ADD_TRACKCURVES = auto()
    MONTAGE_BRANCHING_POINT_REMOVAL = auto()
    BLUEPRINT_ENFORCE_CONST_IN_FUNCTION_OVERRIDES = auto()
    ADD_PIVOT_TO_WIDGET_COMPONENT = auto()
    PAWN_AUTO_POSSESS_AI = auto()
    FTEXT_HISTORY_DATE_TIMEZONE = auto()
    SORT_ACTIVE_BONE_INDICES = auto()
    PERFRAME_MATERIAL_UNIFORM_EXPRESSIONS = auto()
    MIKKTSPACE_IS_DEFAULT = auto()
    LANDSCAPE_GRASS_COOKING = auto()
    FIX_SKEL_VERT_ORIENT_MESH_PARTICLES = auto()
    LANDSCAPE_STATIC_SECTION_OFFSET = auto()
    ADD_MODIFIERS_RUNTIME_GENERATION = auto()
    MATERIAL_MASKED_BLENDMODE_TIDY = auto()
    MERGED_ADD_MODIFIERS_RUNTIME_GENERATION_TO_4_7_DEPRECATED = auto()
    AFTER_MERGED_ADD_MODIFIERS_RUNTIME_GENERATION_TO_4_7_DEPRECATED = auto()
    MERGED_ADD_MODIFIERS_RUNTIME_GENERATION_TO_4_7 = auto()
    AFTER_MERGING_ADD_MODIFIERS_RUNTIME_GENERATION_TO_4_7 = auto()
    SERIALIZE_LANDSCAPE_GRASS_DATA = auto()
    OPTIONALLY_CLEAR_GPU_EMITTERS_ON_INIT = auto()
    SERIALIZE_LANDSCAPE_GRASS_DATA_MATERIAL_GUID = auto()
    BLUEPRINT_GENERATED_CLASS_COMPONENT_TEMPLATES_PUBLIC = auto()
    ACTOR_COMPONENT_CREATION_METHOD = auto()
    K2NODE_EVENT_MEMBER_REFERENCE = auto()
    STRUCT_GUID_IN_PROPERTY_TAG = auto()
    REMOVE_UNUSED_UPOLYS_FROM_UMODEL = auto()
    REBUILD_HIERARCHICAL_INSTANCE_TREES = auto()
    PACKAGE_SUMMARY_HAS_COMPATIBLE_ENGINE_VERSION = auto()
    TRACK_UCS_MODIFIED_PROPERTIES = auto()
    LANDSCAPE_SPLINE_CROSS_LEVEL_MESHES = auto()
    DEPRECATE_USER_WIDGET_DESIGN_SIZE = auto()
    ADD_EDITOR_VIEWS = auto()
    FOLIAGE_WITH_ASSET_OR_CLASS = auto()
    BODYINSTANCE_BINARY_SERIALIZATION = auto()
    SERIALIZE_BLUEPRINT_EVENTGRAPH_FASTCALLS_IN_UFUNCTION = auto()
    INTERPCURVE_SUPPORTS_LOOPING = auto()
    MATERIAL_INSTANCE_BASE_PROPERTY_OVERRIDES_DITHERED_LOD_TRANSITION = auto()
    SERIALIZE_LANDSCAPE_ES2_TEXTURES = auto()
    CONSTRAINT_INSTANCE_MOTOR_FLAGS = auto()
    SERIALIZE_PINTYPE_CONST = auto()
    LIBRARY_CATEGORIES_AS_FTEXT = auto()
    SKIP_DUPLICATE_EXPORTS_ON_SAVE_PACKAGE = auto()
    SERIALIZE_TEXT_IN_PACKAGES = auto()
    ADD_BLEND_MODE_TO_WIDGET_COMPONENT = auto()
    NEW_LIGHTMASS_PRIMITIVE_SETTING = auto()
    REPLACE_SPRING_NOZ_PROPERTY = auto()
    TIGHTLY_PACKED_ENUMS = auto()
    ASSET_IMPORT_DATA_AS_JSON = auto()
    TEXTURE_LEGACY_GAMMA = auto()
    ADDED_NATIVE_SERIALIZATION_FOR_IMMUTABLE_STRUCTURES = auto()
    DEPRECATE_UMG_STYLE_OVERRIDES = auto()
    STATIC_SHADOWMAP_PENUMBRA_SIZE = auto()
    NIAGARA_DATA_OBJECT_DEV_UI_FIX = auto()
    FIXED_DEFAULT_ORIENTATION_OF_WIDGET_COMPONENT = auto()
    REMOVED_MATERIAL_USED_WITH_UI_FLAG = auto()
    CHARACTER_MOVEMENT_ADD_BRAKING_FRICTION = auto()
    BSP_UNDO_FIX = auto()
    DYNAMIC_PARAMETER_DEFAULT_VALUE = auto()
    STATIC_MESH_EXTENDED_BOUNDS = auto()
    ADDED_NON_LINEAR_TRANSITION_BLENDS = auto()
    AO_MATERIAL_MASK = auto()
    NAVIGATION_AGENT_SELECTOR = auto()
    MESH_PARTICLE_COLLISIONS_CONSIDER_PARTICLE_SIZE = auto()
    BUILD_MESH_ADJ_BUFFER_FLAG_EXPOSED = auto()
    MAX_ANGULAR_VELOCITY_DEFAULT = auto()
    APEX_CLOTH_TESSELLATION = auto()
    DECAL_SIZE = auto()
    KEEP_ONLY_PACKAGE_NAMES_IN_STRING_ASSET_REFERENCES_MAP = auto()
    COOKED_ASSETS_IN_EDITOR_SUPPORT = auto()
    DIALOGUE_WAVE_NAMESPACE_AND_CONTEXT_CHANGES = auto()
    MAKE_ROT_RENAME_AND_REORDER = auto()
    K2NODE_VAR_REFERENCEGUIDS = auto()
    SOUND_CONCURRENCY_PACKAGE = auto()
    USERWIDGET_DEFAULT_FOCUSABLE_FALSE = auto()
    BLUEPRINT_CUSTOM_EVENT_CONST_INPUT = auto()
    USE_LOW_PASS_FILTER_FREQ = auto()
    NO_ANIM_BP_CLASS_IN_GAMEPLAY_CODE = auto()
    SCS_STORES_ALLNODES_ARRAY = auto()
    FBX_IMPORT_DATA_RANGE_ENCAPSULATION = auto()
    CAMERA_COMPONENT_ATTACH_TO_ROOT = auto()
    INSTANCED_STEREO_UNIFORM_UPDATE = auto()
    STREAMABLE_TEXTURE_MIN_MAX_DISTANCE = auto()
    INJECT_BLUEPRINT_STRUCT_PIN_CONVERSION_NODES = auto()
    INNER_ARRAY_TAG_INFO = auto()
    FIX_SLOT_NAME_DUPLICATION = auto()
    STREAMABLE_TEXTURE_AABB = auto()
    PROPERTY_GUID_IN_PROPERTY_TAG = auto()
    NAME_HASHES_SERIALIZED = auto()
    INSTANCED_STEREO_UNIFORM_REFACTOR = auto()
    COMPRESSED_SHADER_RESOURCES = auto()
    PRELOAD_DEPENDENCIES_IN_COOKED_EXPORTS = auto()
    TemplateIndex_IN_COOKED_EXPORTS = auto()
    PROPERTY_TAG_SET_MAP_SUPPORT = auto()
    ADDED_SEARCHABLE_NAMES = auto()
    _64BIT_EXPORTMAP_SERIALSIZES = auto()
    SKYLIGHT_MOBILE_IRRADIANCE_MAP = auto()
    ADDED_SWEEP_WHILE_WALKING_FLAG = auto()
    ADDED_SOFT_OBJECT_PATH = auto()
    POINTLIGHT_SOURCE_ORIENTATION = auto()
    ADDED_PACKAGE_SUMMARY_LOCALIZATION_ID = auto()
    FIX_WIDE_STRING_CRC = auto()
    ADDED_PACKAGE_OWNER = auto()
    SKINWEIGHT_PROFILE_DATA_LAYOUT_CHANGES = auto()
    # Added import that can have package different than their outer
    NON_OUTER_PACKAGE_IMPORT = auto()
    # Added DependencyFlags to AssetRegistry
    ASSETREGISTRY_DEPENDENCYFLAGS = auto()
    # Fixed corrupt licensee flag in 4.26 assets
    CORRECT_LICENSEE_FLAG = auto()

    AUTOMATIC_VERSION_PLUS_ONE = auto()
    # AUTOMATIC_VERSION = AUTOMATIC_VERSION_PLUS_ONE.value - 1



from collections import ChainMap, UserString
from dataclasses import dataclass
from typing import List, Optional, TypeVar, Union
from uuid import UUID

from construct import (
    Adapter,
    Array,
    Bytes,
    Check,
    Computed,
    Const,
    Container,
    Error,
    Flag,
    If,
    IfThenElse,
    LazyBound,
    PrefixedArray,
    Rebuild,
    RepeatUntil,
    Switch,
    Tell,
    this,
)
from construct_typed import DataclassMixin, DataclassStruct, csfield

from wacky.basic_types.floats import f32, f64
from wacky.basic_types.guid import GUIDStruct
from wacky.basic_types.integers import i8, i16, i32, i64, u16, u32, u64
from wacky.basic_types.large_bool import LargeBool
from wacky.basic_types.package_index import IndexKind, PackageIndex, PackageIndexStruct
from wacky.basic_types.string import String
from wacky.enums.compression_flags import CompressionFlag, CompressionFlagStruct
from wacky.enums.object_flags import ObjectFlag, ObjectFlagStruct
from wacky.enums.package_flags import PackageFlag, PackageFlagStruct
from wacky.enums.ue4_version import UE4Version
from wacky.hashing.strings import crc32_string_hash, deprecated_string_hash

# from wacky.structs.localization_metadata import GatherableText


@dataclass
class CustomVersion(DataclassMixin):
    key: UUID = csfield(GUIDStruct)
    version: int = csfield(i32)


@dataclass
class Generation(DataclassMixin):
    num_export: int = csfield(u32)
    num_names: int = csfield(u32)


@dataclass
class Version(DataclassMixin):
    major: int = csfield(u16)
    minor: int = csfield(u16)
    patch: int = csfield(u16)
    changelist: int = csfield(u32)
    branch: str = csfield(String)


VersionStruct = DataclassStruct(Version)


@dataclass
class CompressedChunk(DataclassMixin):
    len_compressed: int = csfield(u32)
    ofs_compressed: int = csfield(u32)
    len_uncompressed: int = csfield(u32)
    ofs_uncompressed: int = csfield(u32)


@dataclass
class RawName(DataclassMixin):
    value: str = csfield(String)
    non_case_preserving_hash: Optional[int] = csfield(
        If(
            lambda this: this._root.file_version_ue4
            < UE4Version.NAME_HASHES_SERIALIZED,
            Rebuild(u16, lambda this: deprecated_string_hash(this.value) & 0xFFFF),
        )
    )
    case_preserving_hash: Optional[int] = csfield(
        If(
            lambda this: this._root.file_version_ue4
            < UE4Version.NAME_HASHES_SERIALIZED,
            Rebuild(u16, lambda this: crc32_string_hash(this.value) & 0xFFFF),
        )
    )


class NameAdapter(Adapter):
    def _decode(self, obj: RawName, context, path: str) -> str:
        return obj.value

    def _encode(self, obj: str, context, path) -> RawName:
        return RawName(value=obj)


Name = NameAdapter(DataclassStruct(RawName))


@dataclass
class RawNameReference(DataclassMixin):
    index: int = csfield(u32)
    number: int = csfield(u32)


class ReferencedName(UserString):
    def __init__(self, seq: str, original_reference: RawNameReference):
        super().__init__(seq)
        self.original_reference = original_reference


class NameReferenceAdapter(Adapter):
    def _decode(self, obj: RawNameReference, context, path) -> ReferencedName:
        return ReferencedName(
            context._root.names[obj.index],
            obj,
        )

    def _encode(self, obj: ReferencedName, context, path) -> RawNameReference:
        return obj.original_reference


NameReference = NameReferenceAdapter(DataclassStruct(RawNameReference))


@dataclass
class ImportEntry(DataclassMixin):
    class_package: str = csfield(NameReference)
    class_name: str = csfield(NameReference)
    outer_index: PackageIndex = csfield(PackageIndexStruct)
    object_name: str = csfield(NameReference)


@dataclass
class ExportEntry(DataclassMixin):
    class_index: PackageIndex = csfield(PackageIndexStruct)
    super_index: PackageIndex = csfield(PackageIndexStruct)
    template_index: PackageIndex = csfield(PackageIndexStruct)
    outer_index: PackageIndex = csfield(PackageIndexStruct)
    object_name: str = csfield(NameReference)
    object_flags: ObjectFlag = csfield(ObjectFlagStruct)
    len_serial: int = csfield(u64)
    ofs_serial: int = csfield(u64)
    forced_export: bool = csfield(LargeBool)
    not_for_client: bool = csfield(LargeBool)
    not_for_server: bool = csfield(LargeBool)
    package_guid: UUID = csfield(GUIDStruct)
    package_flags: PackageFlag = csfield(PackageFlagStruct)
    not_always_loaded_for_editor_game: bool = csfield(LargeBool)
    is_asset: bool = csfield(LargeBool)
    first_export_dependency: int = csfield(i32)
    num_serialization_before_serialization_dependencies: int = csfield(u32)
    num_create_before_serialization_dependencies: int = csfield(u32)
    num_serialization_before_create_dependencies: int = csfield(u32)
    num_create_before_create_dependencies: int = csfield(u32)


@dataclass
class PreloadDependencies(DataclassMixin):
    export_map_entry: ExportEntry = csfield(
        Computed(lambda this: this._root.exports_map[this._._index])
    )
    serialization_before_serialization_dependencies: List[PackageIndex] = csfield(
        Array(
            lambda this: this.export_map_entry.num_serialization_before_serialization_dependencies,
            PackageIndexStruct,
        )
    )
    create_before_serialization_dependencies: List[PackageIndex] = csfield(
        Array(
            lambda this: this.export_map_entry.num_create_before_serialization_dependencies,
            PackageIndexStruct,
        )
    )
    serialization_before_create_dependencies: List[PackageIndex] = csfield(
        Array(
            lambda this: this.export_map_entry.num_serialization_before_create_dependencies,
            PackageIndexStruct,
        )
    )
    create_before_create_dependencies: List[PackageIndex] = csfield(
        Array(
            lambda this: this.export_map_entry.num_create_before_create_dependencies,
            PackageIndexStruct,
        )
    )


@dataclass
class MapTag(DataclassMixin):
    key_type: str = csfield(NameReference)
    value_type: str = csfield(NameReference)


@dataclass
class StructTag(DataclassMixin):
    name: str = csfield(NameReference)
    guid: UUID = csfield(GUIDStruct)


TagValue = Union[
    bool,
    str,
    StructTag,
    MapTag,
]


@dataclass
class Tag(DataclassMixin):
    name: str = csfield(NameReference)
    type_: Optional[str] = csfield(
        If(
            lambda this: this.name != "None",
            NameReference,
        )
    )
    size: Optional[int] = csfield(
        If(
            lambda this: this.name != "None",
            i32,
        )
    )
    array_index: Optional[int] = csfield(
        If(
            lambda this: this.name != "None",
            i32,
        )
    )
    value: Optional[TagValue] = csfield(
        If(
            lambda this: this.name != "None",
            Switch(
                lambda this: this.type_,
                {
                    "BoolProperty": Bytes(1),
                    "ByteProperty": NameReference,
                    "EnumProperty": NameReference,
                    "ArrayProperty": NameReference,
                    "SetProperty": NameReference,
                    "MapProperty": DataclassStruct(MapTag),
                    "StructProperty": DataclassStruct(StructTag),
                },
            ),
        )
    )
    has_guid_next: Optional[bool] = csfield(
        If(
            lambda this: this.name != "None",
            Flag,
        )
    )
    guid: Optional[UUID] = csfield(
        If(
            lambda this: this.name != "None" and this.has_guid_next,
            GUIDStruct,
        )
    )


TagStruct = DataclassStruct(Tag)

Value = Union[bytes, bool, int, float, PackageIndex, str, UUID]

NORMAL_VALUE_SWITCH = {
    "ByteProperty": IfThenElse(
        lambda this: this.tag.value == "None",
        Bytes(1),
        NameReference,
    ),
    "BoolProperty": IfThenElse(
        lambda this: not this._root.package_flags & PackageFlag.UnversionedProperties,
        Computed(lambda this: this.tag.value),
        Bytes(1),
    ),
    "IntProperty": i32,
    "FloatProperty": f32,
    "ObjectProperty": PackageIndexStruct,
    "NameProperty": NameReference,
    "DelegateProperty": NameReference,
    "DoubleProperty": f64,
    "ArrayProperty": LazyBound(lambda: DataclassStruct(UArray)),
    # "StructProperty": struct_property # Unused by the DataTable file
    "StrProperty": String,
    # "TextProperty": # TODO
    # "LazyObjectProperty": # TODO
    # "SoftObjectProperty": # TODO
    # "AssetObjectProperty": # TODO
    "UInt64Property": u64,
    "UInt32Property": u32,
    "UInt16Property": u16,
    "Int64Property": i64,
    "Int16Property": i16,
    "Int8Property": i8,
    # "MapProperty": # TODO
    # "SetProperty": # TODO
    "EnumProperty": NameReference,
    "Guid": GUIDStruct,
}

ARRAY_VALUE_SWITCH = ChainMap({"BoolProperty": Bytes(1)}, NORMAL_VALUE_SWITCH)

JSONIFIERS = {
    "ByteProperty": (
        lambda val: str(val)
        if isinstance(val, ReferencedName)
        else int.from_bytes(val, "little")
    ),
    "BoolProperty": (lambda val: bool(val[0]) if isinstance(val, bytes) else bool(val)),
    "IntProperty": int,
    "FloatProperty": float,
    "ObjectProperty": PackageIndex.jsonify,
    "NameProperty": str,
    "DelegateProperty": str,
    "DoubleProperty": float,
    "ArrayProperty": lambda val: val.jsonify(),
    # "StructProperty": struct_property # Unused by the DataTable file
    "StrProperty": str,
    # "TextProperty": # TODO
    # "LazyObjectProperty": # TODO
    # "SoftObjectProperty": # TODO
    # "AssetObjectProperty": # TODO
    "UInt64Property": int,
    "UInt32Property": int,
    "UInt16Property": int,
    "Int64Property": int,
    "Int16Property": int,
    "Int8Property": int,
    # "MapProperty": # TODO
    # "SetProperty": # TODO
    "EnumProperty": str,
    "Guid": str,
}

COMPATIBLE_TYPES = {
    "ByteProperty": [str, int],
    "BoolProperty": [bool],
    "IntProperty": [int],
    "FloatProperty": [float],
    "ObjectProperty": [int],
    "NameProperty": [str],
    "DelegateProperty": [str],
    "DoubleProperty": [float],
    "ArrayProperty": [list],
    # "StructProperty": struct_property # Unused by the DataTable file
    "StrProperty": [str],
    # "TextProperty": # TODO
    # "LazyObjectProperty": # TODO
    # "SoftObjectProperty": # TODO
    # "AssetObjectProperty": # TODO
    "UInt64Property": [int],
    "UInt32Property": [int],
    "UInt16Property": [int],
    "Int64Property": [int],
    "Int16Property": [int],
    "Int8Property": [int],
    # "MapProperty": # TODO
    # "SetProperty": # TODO
    "EnumProperty": [str],
    "Guid": [str],
}

T = TypeVar("T")


def new_if_different(original: T, new: T) -> T:
    if original != new:
        return new
    else:
        return original


def update_byte_property(original: Union[ReferencedName, int], new: Union[str, int]):
    if isinstance(original, ReferencedName):
        if not isinstance(new, str):
            raise ValueError(
                f"Invalid type to update name reference : {new!r} should be a string"
            )
        return update_name_reference(original, new)
    else:
        if type(new) != int:
            raise ValueError(
                f"Invalid type to update byte property : {new!r} should be an int"
            )
        return new_if_different(original, bytes([new]))


def update_bool_property(original: bytes, new: bool) -> bytes:
    if bool(original[0]) == new:
        return original
    else:
        return bytes([new])


def update_package_index(original: PackageIndex, new: int) -> PackageIndex:
    original.update(new)
    return original


def update_array_property(original: "UArray", new: list) -> "UArray":
    original.update(new)
    return original


def update_name_reference(original: ReferencedName, new: str) -> ReferencedName:
    if original != new:
        raise ValueError(
            f"{original!r} is a name reference, updating name references is not supported yet"
        )
    return original


def update_guid(original: UUID, new: str):
    return new_if_different(original, UUID(new))


SPECIAL_UPDATERS = {
    "ByteProperty": update_byte_property,
    "BoolProperty": update_bool_property,
    "ObjectProperty": update_package_index,
    "NameProperty": update_name_reference,
    "DelegateProperty": update_name_reference,
    "ArrayProperty": update_array_property,
    "EnumProperty": update_name_reference,
    "Guid": update_guid,
}


@dataclass
class UArray(DataclassMixin):
    num_items: int = csfield(u32)
    tag: Optional[Tag] = csfield(
        If(
            lambda this: this._.tag.value in ["StructProperty", "ArrayProperty"],
            TagStruct,
        )
    )
    value_type: str = csfield(Computed(this._.tag.value))
    values: List[Value] = csfield(
        Array(
            lambda this: this.num_items,
            Switch(
                lambda this: this.value_type,
                ARRAY_VALUE_SWITCH,
                default=Error,
            ),
        )
    )

    def jsonify(self) -> list:
        return [JSONIFIERS[self.value_type](val) for val in self.values]

    def update(self, new: list):
        raise ValueError("Updating arrays is not yet supported")


@dataclass
class Property(DataclassMixin):
    tag: Tag = csfield(TagStruct)
    value: Optional[Value] = csfield(
        If(
            lambda this: this.tag.name != "None",
            Switch(
                lambda this: this.tag.type_,
                NORMAL_VALUE_SWITCH,
            ),
        )
    )

    def jsonify(self) -> tuple:
        jsonifier = JSONIFIERS.get(self.tag.type_)
        if jsonifier is not None:
            return (str(self.tag.name), jsonifier(self.value))
        else:
            return (str(self.tag.name), None)

    def update(self, change: Value) -> None:
        compatible_types = COMPATIBLE_TYPES.get(self.tag.type_)
        if compatible_types is None:
            if change is None:
                return
            else:
                raise ValueError(
                    f"Invalid value for property {self.tag.name} : "
                    f"should be null but found {change!r}"
                )

        # Can't use isinstance here because bool is a subclass of int
        if type(change) not in compatible_types:
            raise ValueError(
                f"Invalid value for property {self.tag.name} : "
                f"{change!r} is not of type {' or '.join(t.__name__ for t in compatible_types)}"
            )

        special_updater = SPECIAL_UPDATERS.get(self.tag.type_)
        if special_updater is None:
            if self.value != change:
                self.value = change
        else:
            self.value = special_updater(self.value, change)


PropertyStruct = DataclassStruct(Property)
PropertyList = RepeatUntil(
    lambda prop, lst, ctx: prop.tag.name == "None",
    PropertyStruct,
)


@dataclass
class Row(DataclassMixin):
    name: str = csfield(NameReference)
    properties: List[Property] = csfield(PropertyList)

    def jsonify(self) -> tuple:
        return (str(self.name), dict(prop.jsonify() for prop in self.properties[:-1]))

    def update(self, changes: dict) -> None:
        original_props = {prop.tag.name: prop for prop in self.properties[:-1]}
        unknown_props = changes.keys() - original_props.keys()
        if unknown_props:
            raise ValueError(
                f"New data for row {self.name} has unknown properties : {unknown_props}"
            )

        for key, prop_changes in changes.items():
            original_props[key].update(prop_changes)


@dataclass
class DataTable(DataclassMixin):
    properties: List[Property] = csfield(PropertyList)
    has_guid_next: bool = csfield(LargeBool)
    guid: Optional[UUID] = csfield(
        If(
            lambda this: this.has_guid_next
            and not (this._.map_entry.object_flags & ObjectFlag.ClassDefaultObject),
            GUIDStruct,
        )
    )
    rows: List[Row] = csfield(PrefixedArray(u32, DataclassStruct(Row)))

    def jsonify(self) -> dict:
        return {
            "properties": dict(prop.jsonify() for prop in self.properties[:-1]),
            "rows": dict(row.jsonify() for row in self.rows),
        }

    def update(self, changes) -> None:
        original_props = {prop.tag.name: prop for prop in self.properties[:-1]}
        unknown_props = changes["properties"].keys() - original_props.keys()
        if unknown_props:
            raise ValueError(f"New data has unknown properties : {unknown_props}")

        for key, prop_changes in changes["properties"].items():
            original_props[key].update(prop_changes)

        original_rows = {row.name: row for row in self.rows}
        unknown_rows = changes["rows"].keys() - original_rows.keys()
        if unknown_rows:
            raise ValueError(f"New data has unknown rows : {unknown_rows}")

        for name, row_changes in changes["rows"].items():
            original_rows[name].update(row_changes)


def select_actual_type(this: Container) -> str:
    if this.map_entry.class_index.kind == IndexKind.NULL:
        return this.inline_type
    elif this.map_entry.class_index.kind == IndexKind.IMPORT:
        return this._root.imports_map[this.map_entry.class_index.value].object_name
    elif this.map_entry.super_index.kind == IndexKind.IMPORT:
        return this._root.imports_map[this.map_entry.super_index.kind].object_name
    else:
        return this._root.exports_map[this.map_entry.super_index.kind].object_name


@dataclass
class Export(DataclassMixin):
    map_entry: ExportEntry = csfield(
        Computed(lambda this: this._root.exports_map[this._._index])
    )
    computed_offset: int = csfield(Tell)
    inline_type: Optional[str] = csfield(
        If(
            lambda this: this.map_entry.class_index.kind == IndexKind.NULL,
            NameReference,
        )
    )
    type_: str = csfield(Computed(select_actual_type))
    _check_type_is_datatable: None = csfield(Check(this.type_ == "DataTable"))
    value: DataTable = csfield(DataclassStruct(DataTable))

    def jsonify(self) -> dict:
        return self.value.jsonify()

    def update(self, changes: dict) -> None:
        self.value.update(changes)


@dataclass
class Package(DataclassMixin):
    magic: bytes = csfield(Const(bytes([0xC1, 0x83, 0x2A, 0x9E])))
    legacy_file_version: int = csfield(i32)
    version_ue3: Optional[int] = csfield(If(this.legacy_file_version != -4, i32))
    file_version_ue4: int = csfield(i32)
    file_version_licensee_ue4: int = csfield(i32)
    custom_versions: Optional[List[CustomVersion]] = csfield(
        If(
            lambda this: this.legacy_file_version <= -2,
            PrefixedArray(u32, DataclassStruct(CustomVersion)),
        )
    )
    total_header_size: int = csfield(u32)
    folder_name: str = csfield(String)
    package_flags: PackageFlag = csfield(PackageFlagStruct)
    num_names: int = csfield(u32)
    ofs_names: int = csfield(u32)
    # localization_id: Optional[str] = csfield(If(
    #     lambda this: this.file_version_ue4 >= UE4Version.ADDED_PACKAGE_SUMMARY_LOCALIZATION_ID,
    #     String,
    # ))
    num_gatherable_text_data: Optional[int] = csfield(
        If(
            lambda this: this.file_version_ue4 >= UE4Version.SERIALIZE_TEXT_IN_PACKAGES
            or this.file_version_ue4 == 0,
            u32,
        )
    )
    ofs_gatherable_text_data: Optional[int] = csfield(
        If(
            lambda this: this.file_version_ue4 >= UE4Version.SERIALIZE_TEXT_IN_PACKAGES
            or this.file_version_ue4 == 0,
            Const(0, u32),
        )
    )
    num_exports: int = csfield(u32)
    ofs_exports_map: int = csfield(u32)
    num_imports: int = csfield(u32)
    ofs_imports_map: int = csfield(u32)
    ofs_depends_map: int = csfield(u32)
    num_soft_package_references: Optional[int] = csfield(
        If(
            lambda this: this.file_version_ue4
            >= UE4Version.ADD_STRING_ASSET_REFERENCES_MAP
            or this.file_version_ue4 == 0,
            u32,
        )
    )
    ofs_soft_package_references: Optional[int] = csfield(
        If(
            lambda this: this.file_version_ue4
            >= UE4Version.ADD_STRING_ASSET_REFERENCES_MAP
            or this.file_version_ue4 == 0,
            Const(0, u32),
        )
    )
    ofs_searchable_names: Optional[int] = csfield(
        If(
            lambda this: this.file_version_ue4 >= UE4Version.ADDED_SEARCHABLE_NAMES
            or this.file_version_ue4 == 0,
            Const(0, u32),
        )
    )
    ofs_thumbnail_table: int = csfield(Const(0, u32))
    guid: UUID = csfield(GUIDStruct)
    generations: List[Generation] = csfield(
        PrefixedArray(u32, DataclassStruct(Generation))
    )
    saved_by_engine_version: Optional[Version] = csfield(
        If(
            lambda this: this.file_version_ue4 >= UE4Version.ENGINE_VERSION_OBJECT
            or this.file_version_ue4 == 0,
            VersionStruct,
        )
    )
    compatible_with_engine_version: Optional[Version] = csfield(
        If(
            lambda this: this.file_version_ue4
            >= UE4Version.PACKAGE_SUMMARY_HAS_COMPATIBLE_ENGINE_VERSION
            or this.file_version_ue4 == 0,
            VersionStruct,
        )
    )
    compression_flags: CompressionFlag = csfield(CompressionFlagStruct)
    num_compressed_chunks: int = csfield(Const(0, u32))
    # compressed_chunks: List[CompressedChunk] = csfield(PrefixedArray(u32, DataclassStruct(CompressedChunk)))
    source: int = csfield(u32)
    additional_packages_to_cook: List[str] = csfield(PrefixedArray(u32, String))
    num_texture_allocations: Optional[int] = csfield(
        If(
            lambda this: this.legacy_file_version > -7,
            Const(0, u32),
        )
    )
    ofs_asset_registry: int = csfield(u32)
    ofs_bulk_data: int = csfield(u64)
    ofs_world_tile_info: int = csfield(Const(0, u32))
    chunk_ids: List[int] = csfield(PrefixedArray(u32, u32))
    num_preload_dependency: int = csfield(u32)
    ofs_preload_dependency: int = csfield(u32)
    computed_ofs_names: int = csfield(Tell)
    names: List[str] = csfield(Array(this.num_names, Name))
    computed_ofs_gatherable_text_data: int = csfield(Tell)
    # gatherable_text_data: Optional[List[GatherableText]] = csfield(If(
    #     lambda this: not this.package_flags & PackageFlag.FilterEditorOnly,
    #     Array(this.num_gatherable_text_data, DataclassStruct(GatherableText))
    # ))
    computed_ofs_imports_map: int = csfield(Tell)
    imports_map: List[ImportEntry] = csfield(
        Array(this.num_imports, DataclassStruct(ImportEntry))
    )
    computed_ofs_exports_map: int = csfield(Tell)
    exports_map: List[ExportEntry] = csfield(
        Array(this.num_exports, DataclassStruct(ExportEntry))
    )
    computed_ofs_depends_map: int = csfield(Tell)
    depends_map: List[List[int]] = csfield(
        Array(this.num_exports, PrefixedArray(u32, PackageIndexStruct))
    )
    computed_ofs_soft_package_references: int = csfield(Tell)
    # soft_references = ...
    computed_ofs_searchable_names: int = csfield(Tell)
    # searchable_names = ...
    computed_ofs_thumbnail_table: int = csfield(Tell)
    # thumbnails = ...
    computed_ofs_asset_registry: int = csfield(Tell)
    num_assets: int = csfield(Const(0, u32))
    # asset_registry: List[...] = csfield(PrefixedArray(...))
    computed_ofs_world_tile_info: int = csfield(Tell)
    # world_level_info = ...
    computed_ofs_preload_dependency: int = csfield(Tell)
    preload_dependencies: List[PreloadDependencies] = csfield(
        Array(this.num_exports, DataclassStruct(PreloadDependencies))
    )
    computed_total_header_size: int = csfield(Tell)
    exports: List[Export] = csfield(Array(this.num_exports, DataclassStruct(Export)))
    computed_ofs_after_last_export: int = csfield(Tell)
    computed_ofs_bulk_data: int = csfield(Tell)
    magic_at_the_end: bytes = csfield(Const(bytes([0xC1, 0x83, 0x2A, 0x9E])))

    def fix_offsets_and_sizes(self) -> None:
        self.total_header_size = self.computed_total_header_size

        # All the commented lines are for offsets that are hard-coded to zero
        # in the example file
        self.ofs_names = self.computed_ofs_names
        # self.ofs_gatherable_text_data = self.computed_ofs_gatherable_text_data
        self.ofs_exports_map = self.computed_ofs_exports_map
        self.ofs_imports_map = self.computed_ofs_imports_map
        self.ofs_depends_map = self.computed_ofs_depends_map
        # self.ofs_soft_package_references = self.computed_ofs_soft_package_references
        # self.ofs_searchable_names = self.computed_ofs_searchable_names
        # self.ofs_thumbnail_table = self.computed_ofs_thumbnail_table
        self.ofs_asset_registry = self.computed_ofs_asset_registry
        self.ofs_bulk_data = self.computed_ofs_bulk_data
        # self.ofs_world_tile_info = self.computed_ofs_world_tile_info
        self.ofs_preload_dependency = self.computed_ofs_preload_dependency

        if len(self.exports_map) != len(self.exports):
            raise ValueError("ExportsMap and Exports have different lengths")

        num_exports = len(self.exports)
        for i, (map_entry, export) in enumerate(zip(self.exports_map, self.exports)):
            map_entry.ofs_serial = export.computed_offset
            if i + 1 == num_exports:
                map_entry.len_serial = (
                    self.computed_ofs_after_last_export - export.computed_offset
                )
            else:
                map_entry.len_serial = (
                    self.exports[i + 1].computed_offset - export.computed_offset
                )

    def jsonify(self) -> list:
        return [export.jsonify() for export in self.exports]

    def update(self, new_data: list) -> None:
        if len(new_data) != len(self.exports):
            raise ValueError(
                "The new data doesn't have the same number of exports as the old one"
            )

        for changes, export in zip(new_data, self.exports):
            export.update(changes)


PackageStruct = DataclassStruct(Package)



import json
from typing import BinaryIO, TextIO

import click
from ruamel.yaml import YAML

from wacky import dump, jsonify, update


@click.group()
def main() -> None:
    ...


@main.command()
@click.argument("uasset", type=click.File(mode="rb"))
@click.argument("uexp", type=click.File(mode="rb"))
@click.option("--json", "use_json", is_flag=True, help="Output json instead of yaml")
@click.option(
    "-o", "--output", type=click.File(mode="w", encoding="utf-8"), default="-"
)
def unpack(uasset: BinaryIO, uexp: BinaryIO, output: TextIO, use_json: bool) -> None:
    """Convert the file to a more human friendly format"""
    jsonified = jsonify(uasset, uexp)
    if use_json:
        json.dump(jsonified, output, indent=4, ensure_ascii=False)
    else:
        yaml = YAML()
        yaml.dump(jsonified, output)


@main.command()
@click.argument("src_uasset", type=click.File(mode="rb"))
@click.argument("src_uexp", type=click.File(mode="rb"))
@click.argument("new_data", type=click.File(mode="r", encoding="utf-8"))
@click.argument("dst_uasset", type=click.File(mode="wb"))
@click.argument("dst_uexp", type=click.File(mode="wb"))
def repack(
    src_uasset: BinaryIO,
    src_uexp: BinaryIO,
    new_data: TextIO,
    dst_uasset: BinaryIO,
    dst_uexp: BinaryIO,
) -> None:
    """Create new modified .uasset and .uexp files by using SRC_UASSET and
    SRC_UEXP as templates and applying the changes specified by NEW_DATA"""
    yaml = YAML(typ="safe")
    changes = yaml.load(new_data)
    package = update(src_uasset, src_uexp, changes)
    dump(package, dst_uasset, dst_uexp)


if __name__ == "__main__":
    main()



from string import ascii_lowercase, ascii_uppercase

from wacky.hashing import lookup_tables


def deprecated_string_hash(text: str) -> int:
    value = 0
    bytes_ = encode(to_ascii_upper(text))
    for char in bytes_:
        value = ((value >> 8) & 0x00FFFFFF) ^ lookup_tables.DEPRECATED[
            (value ^ char) & 0x000000FF
        ]

    return value


def crc32_string_hash(text: str) -> int:
    value = 0xFFFFFFFF  # 32-bit bitwise inversion of 0
    bytes_ = encode(text)
    char_size = get_char_size(text)
    for i in range(0, len(bytes_), char_size):
        char = int.from_bytes(bytes_[i : i + char_size], "little")
        value = (value >> 8) ^ lookup_tables.SB8[(value ^ char) & 0xFF]
        char >>= 8
        value = (value >> 8) ^ lookup_tables.SB8[(value ^ char) & 0xFF]
        char >>= 8
        value = (value >> 8) ^ lookup_tables.SB8[(value ^ char) & 0xFF]
        char >>= 8
        value = (value >> 8) ^ lookup_tables.SB8[(value ^ char) & 0xFF]

    return value ^ 0xFFFFFFFF  # 32-bit bitwise inversion


def encode(text: str) -> bytes:
    try:
        return text.encode("ascii")
    except UnicodeEncodeError:
        return text.encode("utf-16-le")


UPPER_ASCII_ONLY = str.maketrans(ascii_lowercase, ascii_uppercase)


def to_ascii_upper(text: str) -> str:
    return text.translate(UPPER_ASCII_ONLY)


def get_char_size(text: str) -> int:
    if text.isascii():
        return 1
    else:
        return 2



from enum import IntFlag

from wacky.basic_types.integers import u32
from wacky.enums.adapter import enum_adapter


class CompressionFlag(IntFlag):
    NONE = 0x00

    # Compress with ZLIB - DEPRECATED, USE FNAME
    ZLIB = 0x01
    # Compress with GZIP - DEPRECATED, USE FNAME
    GZIP = 0x02
    # Compress with user defined callbacks - DEPRECATED, USE FNAME
    CUSTOM = 0x04

    # Joint of the previous ones to determine if old flags are being used
    DEPRECATED_FORMAT_FLAGS_MASK = 0x0F

    # Prefer compression that compresses smaller (ONLY VALID FOR COMPRESSION)
    BIAS_MEMORY = 0x10
    # Prefer compression that compresses faster (ONLY VALID FOR COMPRESSION)
    BIAS_SPEED = 0x20
    # Is the source buffer padded out	(ONLY VALID FOR UNCOMPRESS)
    SOURCE_IS_PADDED = 0x80

    # Set of flags that are still allowed
    OPTIONS_FLAGS_MASK = 0xF0


CompressionFlagStruct = enum_adapter(u32, CompressionFlag)



from enum import IntFlag

from wacky.basic_types.integers import u32
from wacky.enums.adapter import enum_adapter


class ObjectFlag(IntFlag):

    NoFlags = 0x00000000

    Public = 0x00000001
    Standalone = 0x00000002
    MarkAsNative = 0x00000004

    Transactional = 0x00000008
    ClassDefaultObject = 0x00000010
    ArchetypeObject = 0x00000020
    Transient = 0x00000040

    MarkAsRootSet = 0x00000080

    TagGarbageTemp = 0x00000100

    NeedInitialization = 0x00000200
    NeedLoad = 0x00000400
    KeepForCooker = 0x00000800
    NeedPostLoad = 0x00001000
    NeedPostLoadSubobjects = 0x00002000
    NewerVersionExists = 0x00004000
    BeginDestroyed = 0x00008000
    FinishDestroyed = 0x00010000

    # Misc. Flags
    BeingRegenerated = 0x00020000
    DefaultSubObject = 0x00040000
    WasLoaded = 0x00080000
    TextExportTransient = 0x00100000
    LoadCompleted = 0x00200000
    InheritableComponentTemplate = 0x00400000
    DuplicateTransient = 0x00800000
    StrongRefOnFrame = 0x01000000
    NonPIEDuplicateTransient = 0x02000000
    Dynamic = 0x04000000
    WillBeLoaded = 0x08000000
    HasExternalPackage = 0x10000000

    # Extras
    Load = (
        Public
        or Standalone
        or Transactional
        or ClassDefaultObject
        or ArchetypeObject
        or DefaultSubObject
        or TextExportTransient
        or InheritableComponentTemplate
        or DuplicateTransient
        or NonPIEDuplicateTransient
    )

    PropagateToSubObjects = Public | ArchetypeObject | Transactional | Transient


ObjectFlagStruct = enum_adapter(u32, ObjectFlag)
